/*
 * diva_wing5_keyboard.cpp
 *
 *  Created on: Sep 28, 2015
 *      Author: michael
 */
#include <diva_wing5_keyboard.h>

#if defined (__USE_LPCOPEN)
#if defined(NO_BOARD_LIB)
#include "chip.h"
#else
#include "board.h"
#endif
#endif
#include "FreeRTOS.h"
#include "task.h"
#include <string.h>
#include <stdlib.h>

#include <cr_section_macros.h>



//!!**!!added
STATIC INLINE void Chip_TIMER_PWMWrite(LPC_TIMER_T *pTMR, uint32_t pwmval)
{
	pTMR->PWMC = pwmval;
}

//!!**!!added
STATIC INLINE uint32_t Chip_TIMER_PWMRead(LPC_TIMER_T *pTMR)
{
	return pTMR->PWMC;
}

enum
{
	KeyboardRowCount = 3,
	KeyboardLedRowCount = 5,
	KeyboardColumnCount = 5,
	KeyboardCellCount = KeyboardRowCount * KeyboardColumnCount,
	KeyboardLedStateCount = 4
};
enum
{
	KeyboardLed_TimerResolution = 1000/*µs*/,
	KeyboardLed_TimerInactiveValue = 10000 /*µs*/
};

static volatile unsigned short ledValues[KeyboardLedStateCount][KeyboardColumnCount][KeyboardLedRowCount];
static unsigned long currentKeyState = 0;
static unsigned char currentKeyCollumn = 0;
static unsigned char currentBlinkState = 0;
static unsigned char currentBlinkTimeout = 0;
static const int columnPorts[KeyboardColumnCount] = {0, 0, 0, 0, 0};
static const int columnPins[KeyboardColumnCount]  = {3, 8, 9, 2, 0};
static const int ledRowPorts[KeyboardLedRowCount] = {0, 1, 1, 1,  0};
static const int ledRowPins[KeyboardLedRowCount]  = {7, 9, 8, 3, 10};
static const int kbdRowPorts[KeyboardRowCount] = {0, 0, 1};
static const int kbdRowPins[KeyboardRowCount]  = {1, 6, 5};

static LPC_TIMER_T * KeyboardTimer(const int _row)
{
	return (_row >= 3) ? LPC_TIMER16_1 : LPC_TIMER16_0;
}

static int8_t KeyboardTimerChannel(const int _row)
{
	return (_row >= 3) ? (_row - 3) : _row;
}

static void handleKeyboardColumn(const int collumn, const int rowsValue)
{
	const unsigned long mask = ~((unsigned long)7 << (collumn*KeyboardRowCount));
	const unsigned long value = rowsValue << (collumn*KeyboardRowCount);
	const unsigned long newKeyState = (currentKeyState & mask) | value;
	if (newKeyState != currentKeyState)
	{
		currentKeyState = newKeyState;
		// keyChanged();
	}
}

static inline void Keyboard_SelectColumn(const int _col)
{
	int i;
	for (i = 0 ; i < KeyboardColumnCount; ++i)
	{
		Chip_GPIO_SetPinState(LPC_GPIO, columnPorts[i], columnPins[i], (_col==i) ? 0 : 1);
	}
}


static inline void Keyboard_SetLedRows(const volatile unsigned short * _rowValues, const int _count)
{
	for (int i=0; i < KeyboardLedRowCount; ++i)
	{
		if (_rowValues[i] >= 10)
		{
			Chip_GPIO_SetPinState(LPC_GPIO, ledRowPorts[i], ledRowPins[i], 1);
		}

		int timerValue = _rowValues[i];
		if ((_rowValues[i] < 10) || (_rowValues[i] >= 990))
		{
			timerValue = KeyboardLed_TimerInactiveValue;
		}
		Chip_TIMER_SetMatch(KeyboardTimer(i), KeyboardTimerChannel(i), timerValue);
	}
}

static inline void Keyboard_ClearLedRows()
{
	for (int row = 0; row < KeyboardLedRowCount; ++row)
	{
		Chip_GPIO_SetPinState(LPC_GPIO, ledRowPorts[row], ledRowPins[row], 0);
		Chip_TIMER_SetMatch(KeyboardTimer(row), KeyboardTimerChannel(row), KeyboardLed_TimerInactiveValue);
	}
}

static inline void Keyboard_ClearLedRow(const int _row)
{
	if (_row < KeyboardLedRowCount)
	{
		Chip_GPIO_SetPinState(LPC_GPIO, ledRowPorts[_row], ledRowPins[_row], 0);
		Chip_TIMER_SetMatch(KeyboardTimer(_row), KeyboardTimerChannel(_row), KeyboardLed_TimerInactiveValue);
	}
}

static inline int Keyboard_ReadRows()
{
	int value = 0;
	int i;
	for (i = 0; i < KeyboardRowCount; ++i)
	{
		if(Chip_GPIO_GetPinState(LPC_GPIO, kbdRowPorts[i], kbdRowPins[i]) == 0)
			value |= 1 << i;
	}
	return value;
}

#if 0
static void manualPollKeyboard()
{
	for (int i = 0; i < KeyboardColumnCount; ++i)
	{
		Keyboard_SelectColumn(i);
		vTaskDelay(configTICK_RATE_HZ / 100);
		handleKeyboardColumn (i, Keyboard_ReadRows());
	}
}
#endif


//LPC_TIMER16_1
extern "C" void TIMER16_1_IRQHandler (void)
{
	if (Chip_TIMER_MatchPending(LPC_TIMER16_1, 0))
	{
		Keyboard_ClearLedRow(3);
		Chip_TIMER_ClearMatch(LPC_TIMER16_1, 0);
	}

	if (Chip_TIMER_MatchPending(LPC_TIMER16_1, 1))
	{
		Keyboard_ClearLedRow(4);
		Chip_TIMER_ClearMatch(LPC_TIMER16_1, 1);
	}
}


// This is a pure C function and must be declared extern "C".
extern "C" void TIMER16_0_IRQHandler (void)
{
	if (Chip_TIMER_MatchPending(LPC_TIMER16_0, 3))
	{
		Chip_TIMER_ClearMatch(LPC_TIMER16_0, 3);
		Chip_TIMER_Disable(LPC_TIMER16_1);

		handleKeyboardColumn (currentKeyCollumn, Keyboard_ReadRows());
		currentKeyCollumn = (currentKeyCollumn + 1) % KeyboardColumnCount;

		Keyboard_ClearLedRows();
		Keyboard_SelectColumn(currentKeyCollumn);
		// select the current state of the blink pattern.
		++currentBlinkTimeout;
		if (currentBlinkTimeout >= 250)
		{
			currentBlinkTimeout = 0;
			currentBlinkState = (currentBlinkState + 1) % KeyboardLedStateCount;
		}

		Keyboard_SetLedRows(&ledValues[currentBlinkState][currentKeyCollumn][0], KeyboardRowCount);

		Chip_TIMER_Enable(LPC_TIMER16_0);
		Chip_TIMER_Reset(LPC_TIMER16_1);
		Chip_TIMER_Enable(LPC_TIMER16_1);
	}
	else
	{
		if (Chip_TIMER_MatchPending(LPC_TIMER16_0, 0))
		{
			Keyboard_ClearLedRow(0);
			Chip_TIMER_ClearMatch(LPC_TIMER16_0, 0);
		}

		if (Chip_TIMER_MatchPending(LPC_TIMER16_0, 1))
		{
			Keyboard_ClearLedRow(1);
			Chip_TIMER_ClearMatch(LPC_TIMER16_0, 1);
		}

		if (Chip_TIMER_MatchPending(LPC_TIMER16_0, 2))
		{
			Keyboard_ClearLedRow(2);
			Chip_TIMER_ClearMatch(LPC_TIMER16_0, 2);
		}
	}
}

/* The definition of IOCON_PIO0_10 in iocon_11xx.h is wrong.
 * defines it as 0x70, but it is 0x68
 */
#define IOCON_PIO0_10_FIXED ((enum CHIP_IOCON_PIO)(0x068 >> 2))

void Keyboard_Initialize(void)
{
	Chip_IOCON_PinMuxSet(LPC_IOCON, IOCON_PIO0_0,  FUNC1|IOCON_DIGMODE_EN); // GPIO instead of RESET
	Chip_IOCON_PinMuxSet(LPC_IOCON, IOCON_PIO0_10_FIXED/*IOCON_PIO0_10*/, FUNC1|IOCON_DIGMODE_EN); // GPIO instead of SWCLK
	Chip_IOCON_PinMuxSet(LPC_IOCON, IOCON_PIO1_3,  FUNC1|IOCON_DIGMODE_EN); // GPIO instead of SWDIO

	for (int s = 0; s < KeyboardLedStateCount; ++s)
	{
		for (int c = 0; c < KeyboardColumnCount; ++c)
		{
			for (int r = 0; r < KeyboardLedRowCount; ++r)
			{
				ledValues[s][c][r] = 0;
			}
		}
	}

	//-- Initialize all Row GPIO's
	for (int row = 0; row < KeyboardLedRowCount; ++row)
	{
		Chip_GPIO_SetPinDIROutput(LPC_GPIO, ledRowPorts[row], ledRowPins[row]);
	}

    for (int i = 0 ; i < KeyboardColumnCount; ++i)
	{
		Chip_GPIO_SetPinDIROutput(LPC_GPIO, columnPorts[i], columnPins[i]);
	}

	for (int i = 0 ; i < KeyboardRowCount; ++i)
	{
		Chip_GPIO_SetPinDIRInput(LPC_GPIO, kbdRowPorts[i], kbdRowPins[i]);
	}

	for (int i=0; i < KeyboardLedRowCount; ++i)
	{
		Chip_GPIO_SetPinState(LPC_GPIO, ledRowPorts[i], ledRowPins[i], 1);
	}

	Chip_TIMER_Init(LPC_TIMER16_0);
	Chip_TIMER_Reset(LPC_TIMER16_0);

	Chip_TIMER_MatchEnableInt(LPC_TIMER16_0, 0);
	Chip_TIMER_MatchEnableInt(LPC_TIMER16_0, 1);
	Chip_TIMER_MatchEnableInt(LPC_TIMER16_0, 2);
	Chip_TIMER_MatchEnableInt(LPC_TIMER16_0, 3);

	Chip_TIMER_ResetOnMatchEnable(LPC_TIMER16_0, 3);
	Chip_TIMER_StopOnMatchEnable(LPC_TIMER16_0, 3);
	Chip_TIMER_PrescaleSet(LPC_TIMER16_0, 72-1); //prescaler:72 = time base 1µs
	Chip_TIMER_SetMatch(LPC_TIMER16_0, 3, KeyboardLed_TimerResolution);	// length in µs (MAT3 )
	Chip_TIMER_SetMatch(LPC_TIMER16_0, 0, KeyboardLed_TimerInactiveValue);	//low in µs
	Chip_TIMER_SetMatch(LPC_TIMER16_0, 1, KeyboardLed_TimerInactiveValue);	//low in µs
	Chip_TIMER_SetMatch(LPC_TIMER16_0, 2, KeyboardLed_TimerInactiveValue);	//low in µs
	Chip_TIMER_PWMWrite(LPC_TIMER16_0, 0); // disable all PWM

	//-- use TIMER_16_1 for more LED's
	Chip_TIMER_Init(LPC_TIMER16_1);
	Chip_TIMER_Reset(LPC_TIMER16_1);

	Chip_TIMER_MatchEnableInt(LPC_TIMER16_1, 0);
	Chip_TIMER_MatchEnableInt(LPC_TIMER16_1, 1);

	Chip_TIMER_PrescaleSet(LPC_TIMER16_1, 72-1); //prescaler:72 = time base 1µs
	Chip_TIMER_SetMatch(LPC_TIMER16_1, 0, KeyboardLed_TimerInactiveValue);	//low in µs
	Chip_TIMER_SetMatch(LPC_TIMER16_1, 1, KeyboardLed_TimerInactiveValue);	//low in µs
	Chip_TIMER_PWMWrite(LPC_TIMER16_1, 0); // disable all PWM

	//-- STart both timers
	Chip_TIMER_Enable(LPC_TIMER16_0); 		//and run...
	NVIC_ClearPendingIRQ(TIMER_16_0_IRQn);
	NVIC_EnableIRQ(TIMER_16_0_IRQn);

	Chip_TIMER_Enable(LPC_TIMER16_1); 		//and run...
	NVIC_ClearPendingIRQ(TIMER_16_1_IRQn);
	NVIC_EnableIRQ(TIMER_16_1_IRQn);
}

void Keyboard_SetLed(const int _column, const int _row, const int _blinkMask, const int _value)
{
	if (_column >= KeyboardColumnCount)
		return;
	if (_row >= KeyboardLedRowCount)
		return;

	for (int i = 0; i < KeyboardLedStateCount; ++i)
	{
		if (_blinkMask & (1<<i))
		{
			ledValues[i][_column][_row] = (_value > 990) ? 990 : _value;
		}
	}
}

unsigned long Keyboard_GetState(void)
{
	return currentKeyState;
}

enum { FaderCount = 5 };

static ADC_CLOCK_SETUP_T ADCSetup;
void Fader_Initialize(void)
{
	Chip_IOCON_PinMuxSet(LPC_IOCON, IOCON_PIO0_11, FUNC2); // AD0
	Chip_IOCON_PinMuxSet(LPC_IOCON, IOCON_PIO1_0, FUNC2);  // AD1
	Chip_IOCON_PinMuxSet(LPC_IOCON, IOCON_PIO1_1, FUNC2);  // AD2
	Chip_IOCON_PinMuxSet(LPC_IOCON, IOCON_PIO1_2, FUNC2);  // AD3
	Chip_IOCON_PinMuxSet(LPC_IOCON, IOCON_PIO1_4, FUNC1);  // AD5

	Chip_ADC_Init(LPC_ADC, &ADCSetup);

	Chip_ADC_EnableChannel(LPC_ADC, ADC_CH0, ENABLE);
	Chip_ADC_EnableChannel(LPC_ADC, ADC_CH1, ENABLE);
	Chip_ADC_EnableChannel(LPC_ADC, ADC_CH2, ENABLE);
	Chip_ADC_EnableChannel(LPC_ADC, ADC_CH3, ENABLE);
	Chip_ADC_EnableChannel(LPC_ADC, ADC_CH5, ENABLE);
	Chip_ADC_SetBurstCmd(LPC_ADC, ENABLE);
	//Chip_ADC_SetStartMode(LPC_ADC, ADC_START_NOW, ADC_TRIGGERMODE_RISING);
}
int Fader_ReadValue(const int _channel)
{
	const unsigned char faderAdcChannel[5] = { ADC_CH0, ADC_CH1, ADC_CH2, ADC_CH3, ADC_CH5 };
	while (Chip_ADC_ReadStatus(LPC_ADC, faderAdcChannel[_channel], ADC_DR_DONE_STAT) != SET) {}
	uint16_t dataADC = 0;
	Chip_ADC_ReadValue(LPC_ADC, faderAdcChannel[_channel], &dataADC);
	return dataADC;
}
